'use strict';
const { CrudService } = require('naf-framework-mongoose-free/lib/service');
const { BusinessError, ErrorCode } = require('naf-core').Error;
const _ = require('lodash');
const assert = require('assert');
const dataToDb = require('../public/dataToDb');
const jwt = require('jsonwebtoken');

// 登陆相关
class LoginService extends CrudService {
  constructor(ctx) {
    super(ctx, 'login');
    this.http = this.ctx.service.util.httpUtil;
    this.spMark = this.ctx.service.special;
    this.prefix = '/db';
    this.time = '30m';
    this.secret = this.config.jwt.secret;

  }

  /**
   * 账号登录,都从这里走
   * 优先使用微信登陆
   * @param {Object} params 表名,直接拿table用
   * @param {Object} body 登陆参数
   * @property phone 电
   * @property password 密码
   * @property openid 微信id
   * @property wx 是否微信登陆
   */
  async login({ table }, { phone, password, openid, wx = false }) {
    assert(table, '缺少登陆角色');
    assert((phone && password) || openid, '缺少登陆信息');
    const query = {};
    if (openid) {
      // TODO微信登录查询
      query.openid = openid;
    } else {
      // 手机号密码登录
      query.phone = phone;
      password = dataToDb.toSecret(password);
    }
    let data = await this.http.$post(`${this.prefix}/fetch`, query, { table });
    if (!data) throw new BusinessError(ErrorCode.USER_NOT_EXIST, '未找到用户');
    if (openid) {
      data = this.spMark.resetDataToFront(data, table);
      // 微信端openid登陆需要token,和token内容,因为没有jwt解密.
      return { token: this.getJwt(table, data), ...this.getTokenData(table, data) };
    }
    // 获取账户密码
    const { password: ap } = data;
    // 账户密码 与 输入密码 对比
    const r = _.isEqual(ap, password);
    if (r) {
      // 密码正确
      data = this.spMark.resetDataToFront(data, table);
      if (!wx) { return this.getJwt(table, data); }
      // 微信登陆需要加密内容和token
      return { token: this.getJwt(table, data), ...this.getTokenData(table, data) };
    }
    throw new BusinessError(ErrorCode.BAD_PASSWORD, '密码错误');

  }

  /**
   * 登陆加密
   * @param {String} table 表名
   * @param {Any} data jwt加密的东西
   */
  getJwt(table, data) {
    const keys = this.getTokenData(table, data);
    return jwt.sign(keys, this.secret);
  }

  /**
   * 筛选jwt加密内容
   * @param {String} table 表名
   * @param {Any} data jwt加密的内容
   */
  getTokenData(table, data) {
    const keys = { table };
    if (data.id) keys.id = data.id;
    return keys;
  }


  /**
   * 验证token
   * @param {Object} token token字符串
   */
  async checkJwt({ token }) {
    if (!token) throw new BusinessError(ErrorCode.ACCESS_DENIED, '缺少秘钥,拒绝访问');
    const errorList = [
      { key: 'jwt expired', word: '秘钥已过期,请重新登陆' },
      { key: 'invalid signature', word: '秘钥错误,请检查秘钥' },
      { key: 'JSON at position', word: '秘钥错误,请检查秘钥' },
      { key: 'invalid token', word: '秘钥错误,请检查秘钥' },
    ];
    try {
      const r = jwt.verify(token, this.secret);
      if (r) return r; // 如果过期将返回false
      return false;
    } catch (e) {
      const { message } = e;
      const r = errorList.find(f => message.includes(f.key));
      if (r) throw new BusinessError(ErrorCode.ACCESS_DENIED, r.word);
      else throw new BusinessError(ErrorCode.ACCESS_DENIED, '秘钥产生位置错误,检测失败');
    }
  }
}

module.exports = LoginService;
