'use strict';
const { CrudService } = require('naf-framework-mongoose-free/lib/service');
const { BusinessError, ErrorCode } = require('naf-core').Error;
const _ = require('lodash');
const assert = require('assert');
const moment = require('moment');
// 微信
class WxService extends CrudService {
  constructor(ctx) {
    super(ctx, 'wx');
    this.userCache = this.ctx.model.WxUserCache;
    this.queryService = this.ctx.service.query;
    this.updateService = this.ctx.service.update;
    const { wx } = this.app.config;
    this.wx = wx;
  }

  // 小程序登录
  async appAuth({ js_code }) {
    const { wxAppConfig } = this.app.config;
    if (!wxAppConfig) return;
    let url = 'https://api.weixin.qq.com/sns/jscode2session';
    // let url = 'http://106.12.161.200/api/position';
    let query = `?js_code=${js_code}`;
    const keys = Object.keys(wxAppConfig);
    for (const key of keys) {
      query = `${query}&${key}=${wxAppConfig[key]}`;
    }
    url = `${url}${query}&grant_type=authorization_code`;
    console.debug('开始请求');
    const res = await this.ctx.curl(url, {
      method: 'get',
      headers: {
        'content-type': 'application/json',
      },
      timeout: 6000,
      dataType: 'json',
    });
    const { openid, unionid } = res.data;
    if (!openid) throw new BusinessError(ErrorCode.BUSINESS, '未获取到openid', res.data);
    return { openid, unionid };
  }

  // 发送模板消息，如果没有openid，查询一下是否
  async sendTemplate(bodyData) {
    let { openid, unionid, templateId, url, topColor = '#ff0000', data } = bodyData;
    const { wx } = this.app.config;
    let wxOpenid = '';
    if (!openid) {
      // 换公众号的openid
      const wxres = await wx.getFollowers();
      if (wxres.data) {
        const wxusers = await wx.batchGetUsers(wxres.data.openid);
        if (wxusers.user_info_list) {
          for (let index = 0; index < wxusers.user_info_list.length; index++) {
            const wxuser = wxusers.user_info_list[index];
            if (wxuser.unionid === unionid) {
              openid = wxuser.openid;
              wxOpenid = wxuser.openid;
              break;
            }
          }
        }
      }
    }
    let res = {};
    // console.log(openid, templateId, url, topColor, data);
    if (openid) {
      res = await wx.sendTemplate(openid, templateId, url, topColor, data);
    } else {
      res.errcode = -1;
      res.errmsg = '发送失败';
      res.msgid = 0;
    }
    res.wxOpenid = wxOpenid;
    return res;
  }

  // 微信关注缓存
  async updateCache() {
    if (!this.app.config.canUseMongoDb) {
      const word = '!!!!!!!!!!mongodb连接测试失败,若需要使用mongodb数据库,需要重启项目进行检测!!!!!!!!!!';
      this.ctx.logger.debug(word);
      return word;
    }
    if (!this.userCache) {
      this.ctx.app.logger.debug('!!!!!!!!!! 未找到 微信关注用户缓存 数据库链接,无法进行 微信关注用户缓存 !!!!!!!!!!');
      this.ctx.ok('!!!!!!!!!! 未找到 微信关注用户缓存 数据库链接,无法进行 微信关注用户缓存 !!!!!!!!!!');
    }
    // 换取每次用户的数量,如果过大.可能会导致js缓存空间被撑死,程序挂掉,所以不用上限10000来搞;可以试着改大些,现在没有那么多关注
    const batchLimit = 1000;
    // 处理数量
    let dealCount = 0;
    // 关注总数
    let wxTotal = 0;
    // 下次开始位置
    let next_openid = null;
    do {
      let arr;
      try {
        // 请求关注列表
        const res = await this.wx.getFollowers(next_openid);
        // 赋值,关注总数
        if (wxTotal === 0) wxTotal = res.data;
        // 赋值下次查询的位置
        next_openid = res.next_openid;
        // 统一接收请求结果的变量
        const result = res.data;
        if (result) {
          // 取出openid列表
          arr = _.get(result, 'openid', []);
          arr = _.chunk(arr, batchLimit);
          // 处理用户,记录数量
          await this.dealUser(arr);
        }
      } catch (error) {
        this.ctx.logger.debug(`微信公众号关注定期拉取:${moment().format('YYYY-MM-DD HH:mm:ss')},处理发生错误`);
      }
      // 拿出来是为了即使里面发生错误.也不会导致无限循环
      dealCount += arr.length;
      // 关注数大于处理数量就继续
    } while (wxTotal > dealCount);
    // 去匹配用户
    await this.userMapping();
  }
  /**
   * 请求每个openid的用户,创建/更新进缓存库
   * @param {Array} arr openid列表
   */
  async dealUser(arr = []) {
    for (const a of arr) {
      const usersRes = await this.wx.batchGetUsers(a);
      const userInfoList = _.get(usersRes, 'user_info_list', []);
      for (const u of userInfoList) {
        const { openid, unionid } = u;
        const obj = { wxopenid: openid, unionid };
        const count = await this.userCache.count({ unionid: obj.unionid });
        if (count) {
          await this.userCache.updateOne({ unionid: obj.unionid }, obj);
        } else {
          await this.userCache.create(obj);
        }
      }
    }
  }

  /**
   * 将没有wxopenid的用户拿出来匹配
   */
  async userMapping() {
    let countTotal = 0;
    let dealTotal = 0;
    let skip = 0;
    const limit = 10;
    const wxUserCache = await this.userCache.find();
    do {
      const query = { skip, limit };
      const res = await this.queryService.index({ table: 'security_guard_base' }, query);
      const { data, total } = res;
      if (countTotal === 0) countTotal = total;
      for (const i of data) {
        const { unionid, id } = i;
        const r = wxUserCache.find(f => f.unionid === unionid);
        if (!r) continue;
        const { wxopenid } = r;
        const q = { id };
        const body = { gopenid: wxopenid };
        this.updateService.index({ table: 'security_guard_base' }, q, body);
      }
      skip += limit;
      dealTotal += data.length;
    } while (countTotal > dealTotal);

  }


  async query() {
    const data = await this.userCache.find();
    const total = await this.userCache.count();
    return { data, total };
  }

}

module.exports = WxService;
